@@ -2,22 +2,27 @@
 #'
 #' @description
 #' Perform various bootstrap process for nested linear mixed effects (LMEs) models including:
-#' parametric, residual, cases, CGR, and REB bootstraps.
+#' parametric, residual, cases, wild, and REB bootstraps.
 #'
 #' @export
 #' @param model The model object you wish to bootstrap.
-#' @param fn A function returning the statistic(s) of interest.
+#' @param .f A function returning the statistic(s) of interest.
 #' @param type A character string indicating the type of bootstrap that is being
 #'    requested. Possible values are \code{"parametric"}, \code{"residual"}, 
-#'    \code{"case"}, \code{"cgr"}, or \code{"reb"} (random effect block bootstrap).
+#'    \code{"case"}, \code{"wild"}, \code{"wild"}, or \code{"reb"} 
+#'    (random effect block bootstrap).
 #' @param B The number of bootstrap resamples.
 #' @param resample A logical vector specifying whether each level of the model 
-#'    should be resampled in the cases bootsrap. The levels should be specified 
+#'    should be resampled in the cases bootstrap. The levels should be specified 
 #'    from the highest level (largest cluster) of the hierarchy to the lowest 
 #'    (observation-level); for example for students within a school, specify the 
 #'    school level first, then the student level.
 #' @param reb_type Specification of what random effect block bootstrap version to
-#' implement. Possible values are \code{0}, \code{1} or \code{2}.
+#'     implement. Possible values are \code{0}, \code{1} or \code{2}.
+#' @param hccme either \code{"hc2"} or \code{"hc3"}, indicating which 
+#'     heteroscedasticity consistent covariance matrix estimator to use.
+#' @param aux.dist either \code{"f1"} or \code{"f2"} indicating which auxiliary 
+#'     distribution to draw the errors from
 #' 
 #' @details
 #' All of the below methods have been implemented for nested linear mixed-effects
@@ -26,18 +31,33 @@
 #' in the help file for that specific function.
 #'   
 #' @return 
-#' The returned value is an object of class "boot", compatible with the \pkg{boot}
-#' package's \code{\link[boot]{boot}} methods.
+#' The returned value is an object of class "lmeresamp". This is a list
+#' with the following elements:
+#' 
+#' \itemize{
+#' \item \code{observed}: the estimated values for the model parameters
+#' \item \code{model}: the fitted model object
+#' \item \code{.f}: the function call
+#' \item \code{replicates}: a \eqn{B \times p} data frame of bootstrap values for each of the p model parameters,
+#' \item \code{stats}: a tibble containing the \]code{observed}, \code{rep.mean} (bootstrap mean), 
+#' \code{se} (bootstrap standard error), and \code{bias} values for each model parameter,
+#' \item \code{B}: the number of bootstrap resamples performed
+#' \item \code{data}: the data with which the model was fit
+#' \item \code{seed}: a vector of randomly generated seeds that are used by the bootstrap
+#' \item \code{type}: the type of bootstrap executed
+#' \item \code{call}: the call to \code{bootstrap()} that the user
+#' \item \code{message}: a list of length \code{B} giving any messages generated during refitting. An entry will be \code{NULL} if no message was generated.
+#' \item \code{warning}: a list of length \code{B} giving any warnings generated during refitting. An entry will be \code{NULL} if no message was generated.
+#' \item \code{error}: a list of length \code{B} giving any errors generated during refitting. An entry will be \code{NULL} if no message was generated.
+#' }
 #' 
 #' @seealso 
 #' \itemize{
 #'   \item \code{\link{parametric_bootstrap}}, \code{\link{resid_bootstrap}},
-#'      \code{\link{case_bootstrap}}, \code{\link{cgr_bootstrap}}, 
-#'      \code{\link{reb_bootstrap}} for more details on a specific bootstrap.
+#'      \code{\link{case_bootstrap}}, \code{\link{reb_bootstrap}}, 
+#'      \code{\link{wild_bootstrap}} for more details on a specific bootstrap.
 #'   \item \code{\link[lme4]{bootMer}} in the \pkg{lme4} package for an 
-#'      implementation of (semi-)parameteric bootstrap for mixed models.
-#'   \item \code{\link[boot]{boot}}, \code{\link[boot]{boot.ci}}, and \code{\link[boot]{plot.boot}} 
-#'      from the \pkg{boot} package.
+#'      implementation of (semi-)parametric bootstrap for mixed models.
 #' }
 #' 
 #' @examples 
@@ -52,40 +72,28 @@
 #'
 #' ## running a parametric bootstrap 
 #' set.seed(1234)
-#' boo1 <- bootstrap(model = vcmodA, fn = mySumm, type = "parametric", B = 100)
+#' boo1 <- bootstrap(model = vcmodA, .f = mySumm, type = "parametric", B = 100)
 #' 
 #' \dontrun{
 #' ## running a cases bootstrap - only resampling the schools
-#' boo2 <- bootstrap(model = vcmodA, fn = mySumm, type = "case", B = 100, resample = c(TRUE, FALSE))
+#' boo2 <- bootstrap(model = vcmodA, .f = mySumm, type = "case", B = 100, resample = c(TRUE, FALSE))
 #' 
 #' ## running a cases bootstrap - resampling the schools and students within the school
-#' boo2 <- bootstrap(model = vcmodA, fn = mySumm, type = "case", B = 100, resample = c(TRUE, FALSE))
+#' boo2 <- bootstrap(model = vcmodA, .f = mySumm, type = "case", B = 100, resample = c(TRUE, FALSE))
 #' 
 #' ## running a semi-parametric bootstrap
-#' boo3 <- bootstrap(model = vcmodA, fn = mySumm, type = "cgr", B = 100)
+#' boo3 <- bootstrap(model = vcmodA, .f = mySumm, type = "cgr", B = 100)
 #' 
 #' ## running a residual bootstrap
-#' boo4 <- bootstrap(model = vcmodA, fn = mySumm, type = "residual", B = 100)
+#' boo4 <- bootstrap(model = vcmodA, .f = mySumm, type = "residual", B = 100)
 #' 
 #' ## running an REB0 bootstrap
-#' boo5 <- bootstrap(model = vcmodA, fn = mySumm, type = "reb", B = 100, reb_typ = 0)
+#' boo5 <- bootstrap(model = vcmodA, .f = mySumm, type = "reb", B = 100, reb_typ = 0)
 #' }
 #' 
 #' ## to print results in a formatted way
-#' requireNamespace("boot") 
-#' boo1
+#' print(boo1)
 #' 
-#' ## you can extract the boostrapped values as a data frame
-#' as.data.frame(boo1$t)
-#' 
-#' ## bootstrap confidence intervals are easily found using 'boot.ci'
-#' ##   warnings about "Some ... intervals may be unstable" go away
-#' ##   for larger bootstrap samples
-#' boot::boot.ci(boo1, index = 1, type=c("norm", "basic", "perc"))
-#' boot::boot.ci(boo1, index = 6, type=c("norm", "basic", "perc"))
-#' 
-#' ## you can also examine the bootstrap samples graphically
-#' plot(boo1, index = 1)
 #'
 #' @references
 #'    Carpenter, J. R., Goldstein, H. and Rasbash, J. (2003) A novel bootstrap 
@@ -107,9 +115,13 @@
 #'    Bates, D., Maechler, M., Bolker, W., Walker, S. (2015).
 #'    Fitting Linear Mixed-Effects Models Using lme4. \emph{Journal of
 #'    Statistical Software}, \bold{67}, 1--48. doi:10.18637/jss.v067.i01.
-bootstrap <- function(model, fn, type, B, resample = NULL, reb_type = NULL) {
-  if(!type %in% c("parametric", "residual", "case", "cgr", "reb"))
-    stop("'type' must be one of 'parametric', 'residual', 'case', 'cgr', or 'reb'")
+#'    
+#'    Modugno, L., & Giannerini, S. (2015). The Wild Bootstrap for 
+#'    Multilevel Models. \emph{Communications in Statistics -- Theory and Methods}, 
+#'    \bold{44}(22), 4812--4825.
+bootstrap <- function(model, .f, type, B, resample = NULL, reb_type = NULL, hccme = NULL, aux.dist = NULL) {
+  if(!type %in% c("parametric", "residual", "case", "wild", "reb"))
+    stop("'type' must be one of 'parametric', 'residual', 'case', 'wild', or 'reb'")
   if(!is.null(reb_type))
     if(!reb_type %in% 0:2) 
       stop("'reb_type' must be either 0, 1, or 2")
@@ -133,18 +145,16 @@ bootstrap <- function(model, fn, type, B, resample = NULL, reb_type = NULL) {
 #' @inheritParams bootstrap
 #' 
 #' @return 
-#' The returned value is an object of class "boot", compatible with the \pkg{boot}
-#' package's \code{\link[boot]{boot}} methods.
+#' The returned value is an object of class "lmeresamp".
 #' 
 #' @seealso 
 #' \itemize{
+#'   \item Examples are given in \code{\link{bootstrap}}
 #'   \item \code{\link{parametric_bootstrap}}, \code{\link{resid_bootstrap}},
-#'      \code{\link{case_bootstrap}}, \code{\link{cgr_bootstrap}}, 
-#'      \code{\link{reb_bootstrap}} for more details on a specific bootstrap.
+#'      \code{\link{case_bootstrap}}, \code{\link{reb_bootstrap}}, 
+#'      \code{\link{wild_bootstrap}} for more details on a specific bootstrap.
 #'   \item \code{\link[lme4]{bootMer}} in the \pkg{lme4} package for an 
-#'      implementation of (semi-)parameteric bootstrap for mixed models.
-#'   \item \code{\link[boot]{boot}}, \code{\link[boot]{boot.ci}}, and \code{\link[boot]{plot.boot}} 
-#'      from the \pkg{boot} package.
+#'      implementation of (semi-)parametric bootstrap for mixed models.
 #' }
 #'
 #' @references
@@ -155,49 +165,10 @@ bootstrap <- function(model, fn, type, B, resample = NULL, reb_type = NULL) {
 #'    Van der Leeden, R., Meijer, E. and Busing F. M. (2008) Resampling multilevel 
 #'    models. In J. de Leeuw and E. Meijer, editors, \emph{Handbook of 
 #'    Multilevel Analysis}, pages 401--433. New York: Springer.
-parametric_bootstrap <- function(model, fn, B) {
+parametric_bootstrap <- function(model, .f, B) {
   UseMethod("parametric_bootstrap", model)
 }
 
-#' @title Residual Bootstrap for Nested LMEs
-#'
-#' @description
-#' Generate residual bootstrap replicates of a statistic for a nested linear 
-#' mixed-effects model.
-#' 
-#' @details
-#' The residual bootstrap resamples the residual quantities from the fitted 
-#' linear mixed-effects model in order to generate bootstrap resamples. That is, 
-#' a random sample, drawn with replacement, is taken from the estimated error terms 
-#' and the EBLUPS (at each level) and the random samples are combined into bootstrap
-#' samples via the fitted model equation.
-#'
-#' @export
-#' @inheritParams bootstrap
-#' 
-#' @return 
-#' The returned value is an object of class "boot", compatible with the \pkg{boot}
-#' package's \code{\link[boot]{boot}} methods.
-#' 
-#' @seealso 
-#' \itemize{
-#'   \item \code{\link{parametric_bootstrap}}, \code{\link{resid_bootstrap}},
-#'      \code{\link{case_bootstrap}}, \code{\link{cgr_bootstrap}}, 
-#'      \code{\link{reb_bootstrap}} for more details on a specific bootstrap.
-#'   \item \code{\link[lme4]{bootMer}} in the \pkg{lme4} package for an 
-#'      implementation of (semi-)parameteric bootstrap for mixed models.
-#'   \item \code{\link[boot]{boot}}, \code{\link[boot]{boot.ci}}, and \code{\link[boot]{plot.boot}} 
-#'      from the \pkg{boot} package.
-#' }
-#'
-#'
-#' @references
-#'    Van der Leeden, R., Meijer, E. and Busing F. M. (2008) Resampling multilevel 
-#'    models. In J. de Leeuw and E. Meijer, editors, \emph{Handbook of 
-#'    Multilevel Analysis}, pages 401--433. New York: Springer.
-resid_bootstrap <- function(model, fn, B) {
-  UseMethod("resid_bootstrap", model)
-}
 
 #' @title Cases Bootstrap for Nested LMEs
 #'
@@ -228,32 +199,30 @@ resid_bootstrap <- function(model, fn, B) {
 #' @inheritParams bootstrap
 #'
 #' @return 
-#' The returned value is an object of class "boot", compatible with the \pkg{boot}
-#' package's \code{\link[boot]{boot}} methods.
+#' The returned value is an object of class "lmeresamp".
 #' 
 #' @seealso 
 #' \itemize{
+#'   \item Examples are given in \code{\link{bootstrap}}
 #'   \item \code{\link{parametric_bootstrap}}, \code{\link{resid_bootstrap}},
-#'      \code{\link{case_bootstrap}}, \code{\link{cgr_bootstrap}}, 
-#'      \code{\link{reb_bootstrap}} for more details on a specific bootstrap.
+#'      \code{\link{case_bootstrap}}, \code{\link{reb_bootstrap}}, 
+#'      \code{\link{wild_bootstrap}} for more details on a specific bootstrap.
 #'   \item \code{\link[lme4]{bootMer}} in the \pkg{lme4} package for an 
-#'      implementation of (semi-)parameteric bootstrap for mixed models.
-#'   \item \code{\link[boot]{boot}}, \code{\link[boot]{boot.ci}}, and \code{\link[boot]{plot.boot}} 
-#'      from the \pkg{boot} package.
+#'      implementation of (semi-)parametric bootstrap for mixed models.
 #' }
 #'
 #' @references
 #'    Van der Leeden, R., Meijer, E. and Busing F. M. (2008) Resampling multilevel 
 #'    models. In J. de Leeuw and E. Meijer, editors, \emph{Handbook of 
 #'    Multilevel Analysis}, pages 401--433. New York: Springer.
-case_bootstrap <- function(model, fn, B, resample) {
+case_bootstrap <- function(model, .f, B, resample) {
   UseMethod("case_bootstrap", model)
 }
 
-#' CGR Bootstrap for Nested LMEs
+#' Residual Bootstrap for Nested LMEs
 #'
 #' @description
-#' Generate semi-parametric bootstrap replicates of a statistic for a nested 
+#' Generate semi-parametric residual bootstrap replicates of a statistic for a nested 
 #' linear mixed-effects model.
 #'
 #' @export
@@ -261,11 +230,12 @@ case_bootstrap <- function(model, fn, B, resample) {
 #' 
 #' @details 
 #' The semi-parametric bootstrap algorithm implemented was outlined by  Carpenter,  
-#' Goldstein and Rasbash (2003). The algorithm is outlined below:
+#' Goldstein and Rasbash (2003), and is referred to as the CGR bootstrap by some. 
+#' The algorithm is outlined below:
 #' \enumerate{
 #'   \item Obtain the parameter estimates from the fitted model and calculate
 #'      the estimated error terms and EBLUPs.
-#'   \item Rescale the error terms and EBLUPs so that the empirical variance of
+#'   \item Center and rescale the error terms and EBLUPs so that the empirical variance of
 #'      these quantities is equal to estimated variance components from the model.
 #'   \item Sample independently with replacement from the rescaled estimated error 
 #'      terms and rescaled EBLUPs.
@@ -275,18 +245,16 @@ case_bootstrap <- function(model, fn, B, resample) {
 #' }
 #'
 #' @return 
-#' The returned value is an object of class "boot", compatible with the \pkg{boot}
-#' package's \code{\link[boot]{boot}} methods.
+#' The returned value is an object of class "lmeresamp".
 #' 
 #' @seealso 
 #' \itemize{
+#'   \item Examples are given in \code{\link{bootstrap}}
 #'   \item \code{\link{parametric_bootstrap}}, \code{\link{resid_bootstrap}},
-#'      \code{\link{case_bootstrap}}, \code{\link{cgr_bootstrap}}, 
-#'      \code{\link{reb_bootstrap}} for more details on a specific bootstrap.
+#'      \code{\link{case_bootstrap}}, \code{\link{reb_bootstrap}}, 
+#'      \code{\link{wild_bootstrap}} for more details on a specific bootstrap.
 #'   \item \code{\link[lme4]{bootMer}} in the \pkg{lme4} package for an 
-#'      implementation of (semi-)parameteric bootstrap for mixed models.
-#'   \item \code{\link[boot]{boot}}, \code{\link[boot]{boot.ci}}, and \code{\link[boot]{plot.boot}} 
-#'      from the \pkg{boot} package.
+#'      implementation of (semi-)parametric bootstrap for mixed models.
 #' }
 #'
 #' @references
@@ -294,8 +262,8 @@ case_bootstrap <- function(model, fn, B, resample) {
 #'    procedure for assessing the relationship between class size and achievement. 
 #'    \emph{Journal of the Royal Statistical Society. Series C (Applied Statistics)}, 
 #'    \bold{52}, 431--443.
-cgr_bootstrap <- function(model, fn, B) {
-  UseMethod("cgr_bootstrap", model)
+resid_bootstrap <- function(model, .f, B) {
+  UseMethod("resid_bootstrap", model)
 }
 
 #' @title REB Bootstrap for Two-Level Nested LMEs
@@ -318,12 +286,14 @@ cgr_bootstrap <- function(model, fn, B) {
 #'      \item predicted random effects \eqn{\tilde{b} = (Z^\prime Z)^{-1} Z^\prime r}
 #'      \item error terms \eqn{\tilde{e} = r - Z \tilde{b}}
 #'   }
-#'   \item Take a simple random sample with replacement of the groups and extract
-#'      the corresponding elements of \eqn{\tilde{b}} and \eqn{tilde{e}}.
+#'   \item Take a simple random sample, with replacement, of the predicted random effects, \eqn{\tilde{b}}.
+#'   \item Draw a simple random sample, with replacement, of the group (cluster) IDs.
+#'       For each sampled cluster, draw a random sample, with replacement, of size 
+#'       \eqn{n_i} from that cluster's vector of error terms, \eqn{\tilde{e}}.
 #'   \item Generate bootstrap samples via the fitted model equation 
 #'      \eqn{y = X \widehat{\beta} + Z \tilde{b} + \tilde{e}}
 #'   \item Refit the model and extract the statistic(s) of interest.
-#'   \item Repeat steps 2-4 B times.
+#'   \item Repeat steps 2-5 B times.
 #' }
 #' 
 #' Variation 1 (\code{type = 1}): 
@@ -338,26 +308,83 @@ cgr_bootstrap <- function(model, fn, B) {
 #' @inheritParams bootstrap
 #'
 #' @return 
-#' The returned value is an object of class "boot", compatible with the \pkg{boot}
-#' package's \code{\link[boot]{boot}} methods.
+#' The returned value is an object of class "lmeresamp".
 #' 
 #' @seealso 
 #' \itemize{
+#'   \item Examples are given in \code{\link{bootstrap}}
 #'   \item \code{\link{parametric_bootstrap}}, \code{\link{resid_bootstrap}},
-#'      \code{\link{case_bootstrap}}, \code{\link{cgr_bootstrap}}, 
-#'      \code{\link{reb_bootstrap}} for more details on a specific bootstrap.
+#'      \code{\link{case_bootstrap}}, \code{\link{reb_bootstrap}}, 
+#'      \code{\link{wild_bootstrap}} for more details on a specific bootstrap.
 #'   \item \code{\link[lme4]{bootMer}} in the \pkg{lme4} package for an 
-#'      implementation of (semi-)parameteric bootstrap for mixed models.
-#'   \item \code{\link[boot]{boot}}, \code{\link[boot]{boot.ci}}, and \code{\link[boot]{plot.boot}} 
-#'      from the \pkg{boot} package.
+#'      implementation of (semi-)parametric bootstrap for mixed models.
 #' }
 #'
 #' @references
 #'    Chambers, R. and Chandra, H. (2013) A random effect block bootstrap for 
 #'    clustered data. \emph{Journal of Computational and Graphical Statistics}, 
 #'    \bold{22}, 452--470.
-reb_bootstrap <- function(model, fn, B, reb_type = 0) {
-  if(!reb_type %in% 0:2) 
-    stop("'reb_type' must be either 0, 1, or 2")
+reb_bootstrap <- function(model, .f, B, reb_type) {
   UseMethod("reb_bootstrap", model)
+}
+
+#' Wild Bootstrap for LME models
+#'
+#' @description
+#' Generate wild bootstrap replicates of a statistic for a 
+#' linear mixed-effects model.
+#'
+#' @export
+#' @inheritParams bootstrap
+#' 
+#' @details 
+#' The wild bootstrap algorithm for LMEs implemented here was outlined by  
+#' Modugno & Giannerini (2015). The algorithm is outlined below:
+#' \enumerate{
+#'   \item Draw a random sample equal to the number of groups (clusters) from
+#'      an auxillary distribution with mean zero and unit variance. 
+#'      Denote these as \eqn{w_1, \ldots, w_g}.
+#'   \item Calculate the selected heteroscedasticity consistent matrix estimator for
+#'      the marginal residuals, \eqn{\tilde{v}_i}
+#'   \item Generate bootstrap responses using the fitted equation: 
+#'      \eqn{y_i^* = X_i \beta + \tilde{v}_i w_j}
+#'   \item Refit the model and extract the statistic(s) of interest.
+#'   \item Repeat steps 2-4 B times.
+#' }
+#'
+#' @return 
+#' The returned value is an object of class "lmeresamp".
+#' 
+#' @seealso 
+#' \itemize{
+#'   \item Examples are given in \code{\link{bootstrap}}
+#'   \item \code{\link{parametric_bootstrap}}, \code{\link{resid_bootstrap}},
+#'      \code{\link{case_bootstrap}}, \code{\link{reb_bootstrap}}, 
+#'      \code{\link{wild_bootstrap}} for more details on a specific bootstrap.
+#'   \item \code{\link[lme4]{bootMer}} in the \pkg{lme4} package for an 
+#'      implementation of (semi-)parametric bootstrap for mixed models.
+#' }
+#'
+#' @references
+#'    Modugno, L., & Giannerini, S. (2015). The Wild Bootstrap for 
+#'    Multilevel Models. \emph{Communications in Statistics -- Theory and Methods}, 
+#'    \bold{44}(22), 4812--4825.
+wild_bootstrap <- function(model, .f, B, hccme, aux.dist) {
+  UseMethod("wild_bootstrap", model)
+}
+
+
+#' Extract parameters from LME models
+#'
+#' @description
+#' A utility function that extracts the fixed effects and variance component
+#' estimates from a fitted \code{merMod} or \code{lme} object.
+#' 
+#' @return 
+#' A named vector of parameters.
+#' 
+#' @export
+#' @param model the fitted model object, either an \code{merMod} or \code{lme} object.
+extract_parameters <- function(model) {
+  UseMethod("extract_parameters", model)
 }
\ No newline at end of file
